#!/usr/bin/env python
# base on Gabriel Genellina http://code.activestate.com/recipes/576755-sorting-big-files-the-python-26-way/
# based on Recipe 466302: Sorting big files the Python 2.4 way
# by Nicolas Lehuen
from __future__ import print_function

import os
import re
from tempfile import gettempdir
from itertools import islice, cycle
from collections import namedtuple
import heapq

Keyed = namedtuple("Keyed", ["key", "obj"])

def merge(key=None, *iterables):
    if key is None:
        for element in heapq.merge(*iterables):
            yield element
    else:
        keyed_iterables = [(Keyed(key(obj), obj) for obj in iterable)
                        for iterable in iterables]
        for element in heapq.merge(*keyed_iterables):
            yield element.obj


def batch_sort(input, output, key=None, buffer_size=10000000, tempdirs=None, uniq=False):
    if tempdirs is None:
        tempdirs = []
    if not tempdirs:
        tempdirs.append(gettempdir())

    chunks = []
    try:
        with open(input, 'rb', 64*1024) as input_file:
            input_iterator = iter(input_file)
            for tempdir in cycle(tempdirs):
                current_chunk = list(islice(input_iterator, buffer_size))
                if not current_chunk:
                    break
                if uniq:
                    current_chunk = list(set(current_chunk))
                current_chunk.sort(key=key)
                output_chunk = open(os.path.join(tempdir, 'sort{0:06d}'.format(len(chunks))), 'w+b', 64*1024)
                chunks.append(output_chunk)
                output_chunk.writelines(current_chunk)
                output_chunk.flush()
                output_chunk.seek(0)
        with open(output, 'wb', 64*1024) as output_file:
            output_file.writelines(merge(key, *chunks))
    finally:
        for chunk in chunks:
            try:
                chunk.close()
                os.remove(chunk.name)
            except Exception:
                pass


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description="Sort a text file by a key in lines of file. This script is compatible with very large files.")
    parser.add_argument(
        'input_file',
        help='Text file to sort'
    )
    parser.add_argument(
        'output_file',
        help='Sorted text file'
    )
    parser.add_argument(
        '-b','--buffer',
        dest='buffer_size',
        default=32000,
        help='''Size of the line buffer. The file to sort is
            divided into chunks of that many lines. Default : 32,000 lines.'''
    )
    parser.add_argument(
        '-k','--key',
        dest='key',
        help='''Python expression used to compute the key for each
            line, "lambda line:" is prepended.\n
            Example : -k "line[5:10]". By default, the whole line is the key.'''
    )
    parser.add_argument(
        '-s','--split',
        dest='split_pattern',
        help='''Python regexp to extract the line key.\n
            Example : -s "^[\w_]+{(?:.+)?}\s+[0-9+.e-]+\s([0-9]+)". By default no regexp work on line chars directly.'''
    )
    parser.add_argument(
        '-t','--tempdir',
        dest='tempdirs',
        action='append',
        default=[],
        help='''Temporary directory to use. You might get performance
            improvements if the temporary directory is not on the same physical
            disk than the input and output directories. You can even try
            providing multiples directories on differents physical disks.
            Use multiple -t options to do that.'''
    )
    parser.add_argument(
        '-u','--uniq',
        dest='uniq',
        action='store_true',
        default=False,
        help='''Sort uniq values.'''
    )
    parser.add_argument(
        '-p','--psyco',
        dest='psyco',
        action='store_true',
        default=False,
        help='''Use Psyco lib to improve execution speed.'''
    )
    options = parser.parse_args()

    lambda_key = None
    if options.split_pattern:
        # If user type --split "^[\w_]+{(?:.+)?}\s+[0-9+.e-]+\s([0-9]+)" --key 1, key will be lasts numbers
        split_pattern = re.compile(options.split_pattern)
        lambda_key = eval('lambda line : (split_pattern.match(line).group({key_index}))'.format(key_index=options.key))
    elif options.key:
        # If user type --key [5:10], key will be extracted as chars between 5 and 10 in line
        lambda_key = eval('lambda line : (line{key_range})'.format(key_range=options.key))

    if options.psyco:
        import psyco
        psyco.full()

    batch_sort(options.input_file, options.output_file, lambda_key, options.buffer_size, options.tempdirs, options.uniq)



import validators
import unicodedata

def _is_free_of_control_characters(text):
    """
    Check if the given string is free of Unicode control characters.

    A control character is any character with a Unicode category of 'Cc'.
    These characters are typically non-printing and could affect the processing
    of text in various environments or applications.
    Full list: https://www.fileformat.info/info/unicode/category/Cc/list.htm

    Parameters:
    - text (str): The string to be checked.

    Returns:
    - bool: True if the string contains no Unicode control characters, False otherwise.
    """
    return all(unicodedata.category(c) != 'Cc' for c in text)


def is_url_valid(new_url):
    """
    Validates the given URL by checking for Unicode control characters and validating its format.

    This function first checks if the input string is free of Unicode control characters using the unicodedata.category method. 
    If it contains control characters, the function prints a message indicating that binary data is being ignored, 
    and returns False. If no control characters are found, it proceeds to check if the string is a valid URL using 
    the validators.url function. If the URL is not valid, it prints a message indicating that an invalid URL is being ignored, 
    and returns False. The function returns True only if the input passes both checks.

    Parameters:
    - new_url (str): The string to be validated as a URL.

    Returns:
    - bool: True if the input is a valid URL and free of Unicode control characters, False otherwise.
    """

    if not _is_free_of_control_characters(new_url):
        print(f"Ignoring binary data: {new_url}")
        return False
    if not validators.url(new_url):
        print(f"Ignoring invalid URL: {new_url}")
        return False
    return True



from urllib.parse import urlunparse, quote_plus, parse_qsl

from p1radup.utils import is_url_valid

class URLProcessor:
    def __init__(self, soft_mode=False):
        self.urls = []
        self.processed_params_by_hostname = {}
        self.soft_mode = soft_mode

    def add_url(self, url):
        """Add a URL to the list of URLs to process."""
        self.urls.append(url)

    def _mark_as_processed(self, hostname, param):
        """Mark a parameter as processed for a given hostname."""
        # Initialize the set if the hostname doesn't exist yet
        if hostname not in self.processed_params_by_hostname:
            self.processed_params_by_hostname[hostname] = set()

        # Add the parameter to the set of processed parameters
        self.processed_params_by_hostname[hostname].add(param)

    def _is_processed(self, hostname, param):
        """Check if a parameter has already been processed for a given hostname."""
        return (hostname in self.processed_params_by_hostname) and (param in self.processed_params_by_hostname[hostname])
    
    def _generate_new_url(self, parsed_url, new_params):
        """
        Generates a new URL with the updated query parameters.
        """
        new_query = "&".join(f"{param}={value}" for param, value in new_params)
        return urlunparse((
            parsed_url.scheme,
            parsed_url.netloc,
            parsed_url.path,
            parsed_url.params,
            new_query,
            parsed_url.fragment
        ))

    def _process_url(self, parsed_url):
        """
        Processes the URL by filtering out seen query parameters and generating a new URL if needed.
        Returns the new URL or None if no new parameters were added.
        """
        if not is_url_valid(parsed_url.geturl()):
            print("Invalid URL couldn't be processed: " + parsed_url.geturl())
            return None

        hostname, path, query = parsed_url.netloc, parsed_url.path, parsed_url.query
        key_for_lookup = (hostname, path) if self.soft_mode else hostname

        # Parse the query string using parse_qsl
        query_params = parse_qsl(query, keep_blank_values=True, strict_parsing=False)

        # Filter out query parameters that have already been processed.
        new_params = []
        for param, value in query_params:
            if not self._is_processed(key_for_lookup, param):
                new_params.append([param, value])
                self._mark_as_processed(key_for_lookup, param)

        # Generate and return the new URL if there are new parameters.
        if new_params:
            # Re-encode the parameters            
            encoded_query_params = [(key, quote_plus(value)) for key, value in new_params]

            new_url = self._generate_new_url(parsed_url, encoded_query_params)
            if not is_url_valid(new_url):
                print("Invalid URL generated from new params: " + new_url)
                return None
            else:
                return new_url
        else:
            return None

    def process_urls(self):
        """
        Process all URLs and return a list of the new URLs.
        """
        results = []

        for parsed_url in self.urls:
            try:
                new_url = self._process_url(parsed_url)
                if new_url:
                    results.append(new_url)
            except Exception as e:
                print(f"Error processing URL: {parsed_url} - {e}")

        return results


#!/usr/bin/env python3

import sys
import os
import datetime
import argparse
import threading
import subprocess
from multiprocessing import Manager
from concurrent.futures import ProcessPoolExecutor, as_completed
from functools import partial
from urllib.parse import urlparse, parse_qsl
from termcolor import colored

from p1radup.sort import batch_sort
from p1radup.url_processor import URLProcessor

def print_banner():
    banner = """
    ███████   ██ ██████   █████  ██████  ██    ██ ██████
    ██   ██ ███ ██   ██ ██   ██ ██   ██ ██    ██ ██   ██
    ███████   ██ ██████  ███████ ██   ██ ██    ██ ██████
    ██       ██ ██   ██ ██   ██ ██   ██ ██    ██ ██
    ██       ██ ██   ██ ██   ██ ██████   ██████  ██

    with <3 by @iambouali and aaznar
    """

    print(colored(banner, 'green'))

def process_chunk(url_processor):
    results = url_processor.process_urls()
    return results

def reader_thread(input_file, chunks_queue, chunk_size, soft_mode):
    url_processor = URLProcessor(soft_mode=soft_mode)
    current_hostname = ''

    current_hostname_processed_params = set()
    current_hostname_newly_seen_params = set()

    for line in input_file:
        try:
            url = line.strip()
            parsed_url = urlparse(url)
        except ValueError:
            print(f"Ignoring invalid URL during read: {url}")
            continue
        
        hostname = (parsed_url.hostname, parsed_url.path) if soft_mode else parsed_url.hostname
        if hostname != current_hostname:
            current_hostname_processed_params = set()
            current_hostname_newly_seen_params = set()
            current_hostname = hostname

        url_processor.add_url(parsed_url)
            
        query = parsed_url.query
        query_params = parse_qsl(query, keep_blank_values=True, strict_parsing=False)

        for param, _ in query_params:
            if param not in current_hostname_processed_params:
                current_hostname_newly_seen_params.add(param)

        if len(url_processor.urls) >= chunk_size:
            current_hostname_processed_params.update(current_hostname_newly_seen_params)
            url_processor.processed_params_by_hostname[hostname] = current_hostname_processed_params
            chunks_queue.put(url_processor)

            current_hostname_newly_seen_params = set()
            url_processor = URLProcessor(soft_mode=soft_mode)

    if url_processor.urls:  # Ensure the last chunk is added
        chunks_queue.put(url_processor)

    # Signal that reading is done by adding None
    chunks_queue.put(None)

def worker(chunks_queue, output_file=None, output_file_lock=None):
    try:
        while True:
            chunk = chunks_queue.get()
            if chunk is None:  # End signal
                chunks_queue.put(None)  # Propagate the end signal for other workers
                break
            results = process_chunk(chunk)
            if output_file:
                with output_file_lock:
                    with open(output_file, 'a', encoding='utf-8', errors='ignore') as file:
                        for result in results:
                            file.write(result + '\n')
            else:
                for result in results:
                    print(result)
    except Exception as e:
        print(f"Worker error: {e}")

def process_urls_with_pool(input_file, output_file, soft_mode, chunk_size, num_workers):
    m = Manager()
    chunks_queue = m.Queue(num_workers * 2)
    output_file_lock = m.Lock()

    # Start the reader thread
    reader = threading.Thread(target=reader_thread, args=(input_file, chunks_queue, chunk_size, soft_mode))
    reader.start()

    # Create a pool of worker processes
    with ProcessPoolExecutor(max_workers=num_workers) as executor:
        # Use partial to create a function with fixed arguments (output_file, output_file_lock)
        worker_partial = partial(worker, output_file=output_file, output_file_lock=output_file_lock)
        futures = [executor.submit(worker_partial, chunks_queue) for _ in range(num_workers)]

        # Wait for all futures to complete. This loop is not strictly necessary in this setup,
        # but it's useful if you want to handle exceptions or results from workers.
        for future in as_completed(futures):
            try:
                future.result()  # If a worker raised an exception, it will be re-raised here.
            except Exception as e:
                print(f"Worker error: {e}")

    # Wait for the reader thread to finish
    reader.join()

def sort_and_save_input_lines(input_source, use_gnu_sort=False):
    timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")

    if not input_source:
        output_filename = f"/tmp/input_{timestamp}.txt"

        with open(output_filename, 'a', buffering=-1, encoding='utf-8') as file:
            for line in sys.stdin:
                file.write(line)

    if input_source:
        input = input_source
        filename = os.path.splitext(os.path.basename(input_source))[0]
        output_filename = f"/tmp/{filename}_sorted_{timestamp}.txt"
        
        if use_gnu_sort:
            with open(output_filename, 'w', encoding='utf-8') as output_file:
                subprocess.run(['sort', input], stdout=output_file)
        else:
            batch_sort(input, output_filename)

    return output_filename

def main():
    print_banner()

    parser = argparse.ArgumentParser(description='Process URLs and remove duplicate query parameters.')
    parser.add_argument('-i', '--input', help='Input file path')
    parser.add_argument('-o', '--output', help='Output file path')
    parser.add_argument('-s', '--soft-mode', help='Enable soft mode to preserve duplicates in different paths and the same hostname', action='store_true')
    parser.add_argument('-c', '--chunk-size', type=int, default=50000, help='The size of each chunk of URLs to process at a time')
    parser.add_argument('-w', '--num-workers', type=int, default=4, help='The number of worker processes (an additional thread is used for reading the input file)')
    parser.add_argument('-gs', '--gnu-sort', help='Use GNU sort instead of custom sorting', action='store_true')

    args = parser.parse_args()

    # Delete output file if it already exists
    if args.output is not None and os.path.exists(args.output):
        os.remove(args.output)

    sorted_filename = sort_and_save_input_lines(args.input, args.gnu_sort)
    sorted_file = open(sorted_filename, 'r', encoding='utf-8')

    process_urls_with_pool(sorted_file, args.output, args.soft_mode, args.chunk_size, args.num_workers)

    # After processing URLs
    sorted_file.close()  # Close the file handle

    # Attempt to delete the sorted file
    if os.path.exists(sorted_filename):
        try:
            os.remove(sorted_filename)
        except Exception as e:
            print(f"Error deleting sorted file: {e}")
    else:
        print(f"Sorted file {sorted_filename} does not exist.")

if __name__ == '__main__':
    main()
